/*
 * Prologix GPIB-ETHERNET/GPIB-USB controller library
 *
 * 2009-09-30 Kouichi Hirabayashi
 */
#include <stdio.h>
#include <signal.h>
#include <stdarg.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>

#define PORT_NO		1234

#define TIMEOUT		5

#define USBDEV		0
#define LANDEV		1

#define GB_IDLE		0	/* idle */
#define GB_NCNT		1	/* non-controller */
#define GB_CNTR		2	/* non-system controller */
#define GB_SCNT		3	/* system controller */
#define GB_LONY		4	/* lsten only */
#define GB_TOLY		5	/* talk only */

static void onalarm(int sig);
static void error(char *s, ...);

#ifndef OLDSIGNAL
struct sigaction saalrm, osaalrm;
#endif
static int gbtype;
static int gbfd;
static int gbtout;
static int gbst;
static int gbmode;
static void (*usrfun)(void);

gbopn(char *port)
{
  char tmps[32];

  gbst = 0;
  gbtout = TIMEOUT;
  if (strncmp(port, "/dev/", 4) == 0) {
	if (usbOpen(port, B19200) < 0)
		error("can't open %s", port);
  }
  else {
	if (lanOpen(port, PORT_NO) < 0)
		error("can't open %s", port);
  }
#ifdef E5810_COMPAT
  if (serPuts(gbfd, "++mode 1\r", 0) < 0)
	error("gbopn(): %s error.", "++mode 1");
  if (serPuts(gbfd, "++auto 0\r", 0) < 0)
	error("gbopn(): %s error.", "++auto 0");
  sprintf(tmps, "++read_tmo_ms 3000\r");
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbred(): %s error.", "++read_tmo_ms 3000");
  gbifc(gbfd);
  gbmode = GB_SCNT;
#endif
#ifdef OLDSIGNAL
  signal(SIGALRM, onalarm);
#else
  saalrm.sa_handler = onalarm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
#endif
  return 0;
}

gbifc()
{
  char tmps[32];

  sprintf(tmps, "++ifc\r");
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbifc() failed.", (char *)0);
  return 0;
}

gbclr(int addr)
{
  char tmps[32];

  sprintf(tmps, "++addr %d\r", addr);
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbclr(): ++addr %s error.", tmps);
  if (serPuts(gbfd, "++clr\r", 0) < 0)
	error("gbclr() failed.", (char *)0);
  return 0;
}

gblcl(int addr)
{
  char tmps[32];

  sprintf(tmps, "++addr %d\r", addr);
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gblcl(): ++addr %s error.", tmps);
  if (serPuts(gbfd, "++loc\r", 0) < 0)
	error("gblcl() failed.", (char *)0);
  return 0;
}

gbtrg(int addr)
{
  char tmps[32];

  sprintf(tmps, "++addr %d\r", addr);
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbtrg(): ++addr %s error.", tmps);
  sprintf(tmps, "++trg %d\r", addr);
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbtrg() failed.", (char *)0);
  return 0;
}

gbspol(int addr)
{
  int stb;
  char tmps[32];

  sprintf(tmps, "++spoll %d\r", addr);
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbspol() ++spoll %d error.", addr);
  if (serGets(gbfd, tmps, sizeof(tmps), 2) < 0)
	error("gbspol() read error.", (char *)0);
  sscanf(tmps, "%d", &stb);
  return stb;
}

gbwrt(int addr, char *buf, int siz, int dlm)
{
  static int dt[] = { 3, 3, 0, 1 };
  static char *lbuf;
  static int lbufsiz;
  int i, n, m;
  char *s, *t, tmps[32];

  if (gbmode == GB_IDLE)
	gbcmod(GB_SCNT);
  if (siz == 0)
	siz = strlen(buf);
  if (siz == 0)
	return(0);
  if (addr > -1) {
	if (addr < 100)
		sprintf(tmps, "++addr %d\r", addr & 0x1F);
	else
		sprintf(tmps, "++addr %d %d\r", (addr / 100) & 0x1F,
			addr & 0xFF);
	if (serPuts(gbfd, tmps, 0) < 0)
		error("gbwrt() ++addr %s error.", tmps);
  }
  sprintf(tmps, "++eos %d\r", dt[dlm & 0x03]);
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbwrt(): ++eos %d error.", dt[dlm & 0x03]);
  sprintf(tmps, "++eoi %d\r", (dlm & 1) ? 1 : 0);
  m = 1;	/* \r */
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbwrt(): ++eoi %d error.", (dlm & 1) ? 1 : 0);
  if (gbmode == GB_NCNT) {
	if (serPuts(gbfd, "++mode 0\r", 0) < 0)
		error("gbwrt(): ++mode 0 error.", (char *)0);
  }
  if (siz == 0)
	siz = strlen(buf);
  for (n = i = 0, s = buf; i < siz; i++)
	if (strchr("\053\033\r\n", *s++ & 0xFF) != (char *)NULL)
		n++;

  i = siz + n + m;
  if (lbufsiz == 0) {
	if ((lbuf = (char *)malloc(i)) == (char *)NULL)
		error("memory over.", (char *)0);
	lbufsiz = i;
  }
  else if (i > lbufsiz) {
	if ((lbuf = (char *)realloc(lbuf, i)) == (char *)NULL)
		error("memory over.", (char *)0);
	lbufsiz = i;
  }
  if (n > 0) {
	for (i = 0, s = buf, t = lbuf; i < siz; i++) {
		if (strchr("\053\033\r\n", *s) != (char *)NULL)
			*t++ = '\033';
		*t++ = *s++;
	}
  }
  else {
	for (i = 0, s = buf, t = lbuf; i < siz; i++)
		*t++ = *s++;
  }
  *t++ = '\r';
  if (serPuts(gbfd, lbuf, siz + n + m) < 0)
	error("gbwrt() error.", (char *)0);
  return 0;
}

gbred(int addr, char *buf, int siz, int dlm)
{
  static int dt[] = { 0, 0, 10, 13 };
  int n;
  char tmps[32];

  if (gbmode == GB_IDLE)
	gbcmod(GB_SCNT);
  if (addr > -1) {
	if (addr < 100)
		sprintf(tmps, "++addr %d\r", addr & 0x1F);
	else
		sprintf(tmps, "++addr %d %d\r", (addr / 100) & 0x1F,
			addr & 0xFF);
	if (serPuts(gbfd, tmps, 0) < 0)
		error("gbwrt() %s error.", tmps);
  }
  if (dlm == 0) {
	if (serPuts(gbfd, "++eot_enable 0\r", 0) < 0)
		error("gbred(): %s error.", "++eot_enable 0");
	sprintf(tmps, "++read\r");
  }
  else if (dlm == 1) {
	if (serPuts(gbfd, "++eot_char 42\r", 0) < 0)
		error("gbred(): %s error.", "++eot_char 42");
	if (serPuts(gbfd, "++eot_enable 1\r", 0) < 0)
		error("gbred(): %s error.", "++eot_enable 1");
	sprintf(tmps, "++read eoi\r");
  }
  else {
	if (serPuts(gbfd, "++eot_enable 0\r", 0) < 0)
		error("gbred(): %s error.", "++eot_enable 0");
	sprintf(tmps, "++read %d\r", dt[dlm & 0x03]);
  }
  if (gbmode == GB_SCNT || gbmode == GB_CNTR) {
	if (serPuts(gbfd, tmps, 0) < 0)
		error("gbred(): %s error.", tmps);
  }
  gbst = -1;
  if ((n = serGets(gbfd, buf, siz, dlm)) < 0)
	error("gbred() read error.", (char *)0);
 return n;
}

gbwait(mask)
{
  static time_t clk0, clk;
  static int r0;
  int i, r;
  char tmps[32];

  for (r = 0; !(r & 0x40); ) {
	time(&clk);
	if (clk - clk0 <= 0)
		r = r0;
	else {
		if (serPuts(gbfd, "++srq\r", 0) < 0)
			error("gbwait() error.", (char *)0);
		  if (serGets(gbfd, tmps, sizeof(tmps), 2) < 0)
			error("gbwait() read error.", (char *)0);
		r = (*tmps == '0') ? 0 : 0x40;
	}
	clk0 = clk;
	if (mask <= 0)
		break;
	usleep(100000);
  }
  if (mask < 0)
	r |= gbst;
  return r;
}

gbcls()
{
  return (gbtype == USBDEV) ? usbClose() : lanClose();
}

gblost(int sec)
{
  gbtout = sec;
  return 0;
}

gbver()
{
  char tmps[64];
  int n;

  if (serPuts(gbfd, "++ver\r", 0) < 0)
	error("gbver(): %s error.", "++var");
  if ((n = serGets(gbfd, tmps, sizeof(tmps), 2)) < 0)
	error("gbver() read error.", (char *)0);
  printf("%s\n", tmps);
  return 0;
}

gbcmod(int mod)
{
  char tmps[64];

#ifndef E5810_COMPAT
  if (mod == GB_NCNT) {	/* NCNT */
	if (serPuts(gbfd, "++mode 0\r", 0) < 0)
		error("gbcmod(): %s error.", "++mode 0");
	gbmode = GB_NCNT;
  }
  else if (mod == GB_CNTR || mod == GB_SCNT) {	/* CNTR, SCNT */
	if (serPuts(gbfd, "++mode 1\r", 0) < 0)
		error("gbcmod(): %s error.", "++mode 1");
	if (serPuts(gbfd, "++auto 0\r", 0) < 0)
		error("gbcmod(): %s error.", "++auto 0");
	sprintf(tmps, "++read_tmo_ms 3000\r");
	if (serPuts(gbfd, tmps, 0) < 0)
		error("gbcmod(): %s error.", "++read_tmo_ms 3000");
	gbifc(gbfd);
	gbmode = GB_SCNT;
  }
  else {
	fprintf(stderr, "Warning: listen only ot talk only mode not supported.\n");
  }
#else
  if (mod != 3)
	fprintf(stderr, "gbcmod(%d) is not supported.\n", mod);
#endif
  return 0;
}

gbrqs(int stb)
{
  char tmps[64];
  int r;

  sprintf(tmps, "++status %d\r", stb & 0xFF);
  if (serPuts(gbfd, tmps, 0) < 0)
	error("gbrqs(): %s error.", tmps);
  if (serPuts(gbfd, "++status\r", 0) < 0)
	error("gbrqs(): %s error.", "++status");
  if (!(stb & 0x40))
	return 0;
  for (;;) {
	if (serGets(gbfd, tmps, sizeof(tmps), 2) < 0)
		error("gbrqs() timeout.", (char *)0);
	sscanf(tmps, "%d", &r);
	if (!r)
		break;
	sleep(1);
	if (serPuts(gbfd, "++status\r", 0) < 0)
		error("gbrqs(): %s error.", "++status");
  };
  return 0;
}

gbabort()
{
  return gbifc();
}

gbadr(int ad)
{
  char tmps[64];

  if (gbmode == GB_NCNT) {
	sprintf(tmps, "++addr %d\r", ad & 0x1F);
	if (serPuts(gbfd, tmps, 0) < 0)
		error("gbadr() %s error.", tmps);
  }
  return 0;
}

gbrmt(int f)
{
  fprintf(stderr, "gbrmt() not supported.\n");
  return 0;
}

gbmsg(char *msg)
{
  fprintf(stderr, "gbmsg() not supported.\n");
  return 0;
}

gbppol(void)
{
  fprintf(stderr, "gbppol() not supported.\n");
  return 0;
}

gbppc(int ad, int data)
{
  fprintf(stderr, "gbppc() not supported.\n");
  return 0;
}

gbppu(int ad)
{
  fprintf(stderr, "gbppu() not supported.\n");
  return 0;
}

gblost2()
{
  fprintf(stderr, "Warming: gblost2() not supported.\n");
  return 0;
}

gblock()
{
  fprintf(stderr, "Warming: gblock() not supported.\n");
  return 0;
}

gbunlock()
{
  fprintf(stderr, "Warming: gbunlock() not supported.\n");
  return 0;
}

gberror()
{
  fprintf(stderr, "Warming: gberror() not supported.\n");
  return 0;
}

gbint(void *fun)
{
  usrfun = fun;
  return 0;
}

usbOpen(port, speed) char *port;
{
  struct termios tios;

  /* reopen to enabme c_cc[VMIN] and c_cc[VTIME] */
  if ((gbfd = open(port, O_RDWR)) == -1)
	error("can't open %s", port);
  tios.c_iflag = 0;
  tios.c_oflag = 0;
  tios.c_cflag = CLOCAL|CREAD|CS8|HUPCL|CRTSCTS;
  tios.c_lflag = 0;
  tios.c_cc[VEOL] = '\r';
  tios.c_cc[VMIN] = 0;
  tios.c_cc[VTIME] = 10;
  if (cfsetospeed(&tios, speed) == -1)
	error("cfsetospeed() failed.", (char *)0);
  if (cfsetispeed(&tios, speed) == -1)
	error("cfsetispeed() failed.", (char *)0);
  if (tcsetattr(gbfd, TCSANOW, &tios) == -1)
	error("ioctl(TCSANOW) failed", (char *)0);
  gbtype = USBDEV;
  return 0;
}

usbClose()
{
  return close(gbfd);
}

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

static struct sockaddr_in sv_addr;
static struct hostent *shost;

static char srv_addr[80];

lanOpen(char *ipadr, int port)
{
  strcpy(srv_addr, ipadr);      /* IP Address */
  alarm(gbtout);
  if ((gbfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	error("socket() failed", (char *)0);
  if ((shost = gethostbyname(srv_addr)) == NULL)
	error("gethostbyname() failed", (char *)0);
  memset((char *)&sv_addr, 0, sizeof(sv_addr));
  sv_addr.sin_family = AF_INET;
  sv_addr.sin_port = htons(port);
  memcpy((char *)&sv_addr.sin_addr, (char *)shost->h_addr,
	shost->h_length);
  if (connect(gbfd, (struct sockaddr *)&sv_addr, sizeof(sv_addr)) < 0) {
	return -1;
  }
  alarm(0);
  gbtype = LANDEV;
  return 0;
}

lanClose()
{
  if (shutdown(gbfd, 2) < 0)
	error("shutdown() failed", (char *)0);
  return close(gbfd);
}

serPuts(fd, s, n) char *s;
{
  if (n == 0)
	n = strlen(s);
  if (write(fd, s, n) != n)
		return -1;
  return 0;
}

serGets(fd, buf, cnt, dlm) char *buf;
{
  static int dt[] = { 0, 42, 10, 13 };
  int i, k, n, r, st = -1;
  char *s, *t;

  alarm(gbtout);
  for (r = cnt, s = buf, n = 0; r > 0; ) {
	if ((k = read(fd, s, r)) < 0) {
		if (dlm > 0)
			return -1;
		fprintf(stderr, "Warning: gbred() timeout in %d sec.\n", gbtout);
		goto g1;
	}
	if (k == 0) {
		usleep(10000);
		continue;
	}
	n += k;
	r -= k;
	t = s;
	s += k;
	if (dlm > 0 && (s[-1] & 0xFF) == dt[dlm])
		goto g1;
  }
g1:
  alarm(0);
  if (r > 0) {
	if (dlm == 2 && s[-2] == 0x0D) {
		s[-2] = '\0';
		n -= 2;
		st = 0x02;
	}
	else if ((dlm & 0x01) == 1 && s[-1] == dt[dlm]) {
		s[-1] = '\0';
		n -= 1;
		st = (dlm == 1) ? 0x04 : 0x02;
	}
	else
		*s = '\0';
  }
  else
	st = 0x01;
  if (gbst == -1)
	gbst = st;
  return n;
}

static void
onalarm(int sig)
{
#ifdef OLDSIGNAL
  signal(SIGALRM, onalarm);
#else
  saalrm.sa_handler = onalarm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
#endif
}

static void
error(char *fmt, ...)
{
  va_list argp;

  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
  fprintf(stderr, "\n");
  exit(1);
}
